#include "stdafx.h"

#define DllExport __declspec(dllexport)
#include "mcli.h"

extern "C"
{
	#define MULTI_THREAD_RECEIVER
	#define pthread_setcancelstate(x, y);
	#define pthread_mutex_t int
	#define STATIC static
	#define _SOTYPE const char *
	#define socklen_t int
	#define SA SOCKADDR
	#define uint8_t unsigned char
	#define poll WSAPoll
	static void err(const char *fmt, ...) { va_list args; va_start(args, fmt); fprintf(stderr, "err: "); vfprintf(stderr, fmt, args); }
	static void warn(const char *fmt, ...) { va_list args; va_start(args, fmt); fprintf(stderr, "wrn: "); vfprintf(stderr, fmt, args); }
	static void dbg(const char *fmt, ...) { va_list args; va_start(args, fmt); fprintf(stderr, "dbg: "); vfprintf(stdout, fmt, args); }
	#include "mcast.h"
	#include "mcast.c"
	#pragma comment(lib, "ws2_32")
	#pragma comment(lib, "iphlpapi")
	const char *simFile = NULL;
	bool simRepeat = true;
	bool simSlow = false;
	void setSimFile(const char *f){ simFile=f; }
	static int random(int max) { return (int)((double)rand() / (double)(RAND_MAX + 1) * max); }
	int udp_read_test(UDPContext * s, uint8_t * buf, int size, int timeout, struct sockaddr_storage *from)
	{
		if (simFile)
		{
			static FILE *f = NULL;
			if (f!=NULL && (simRepeat && feof(f))) { fclose(f); f=NULL; }
			if (f == NULL) f = fopen(simFile, "r");
			if (!simSlow) return f==NULL?-1:fread(buf, 1, size, f);
			if (random(100) > 80) { Sleep(50); return 0; }
			return f==NULL?-1:fread(buf, 1, random(80), f);
	}
		return udp_read(s, buf, size, timeout, from);
	}
	#define udp_read udp_read_test
}

string format(const char *fmt, ...) { va_list args; char buf[64*1024]; va_start(args, fmt); vsprintf_s(buf, fmt, args); return buf; }

Netceiver::Netceiver(const string &ifname, int port, char mcg[][1024], int num_mcg) : m_client(NULL), m_mldmcgs(0), m_subscribers(0)
{
	m_buf = new char[1024*1024];
	m_udp = new unsigned char[UDP_TX_BUF_SIZE+1];
	m_ifname = ifname;
	m_port = port;
	char _mcg[10][1024];
	if (mcg == NULL) { strcpy(_mcg[0], "ff18:5100::"); mcg = _mcg; num_mcg = 1; }
	m_num_mcg = num_mcg;
	for (int i=0; i<m_num_mcg; i++) strcpy(m_mcg[i], mcg[i]);
	m_client = client_udp_open_host(mcg[0], port, ifname=="*"?NULL:ifname.c_str());
	if (m_client)
	{
		struct sockaddr_in6 addr;
		addr.sin6_family = AF_INET6;
		addr.sin6_port = htons(port);
		for (int i=1; i<m_num_mcg; i++)
		{
			inet_pton(AF_INET6, mcg[i], &addr.sin6_addr);
			if (udp_ipv6_join_multicast_group(((UDPContext *)m_client)->udp_fd, ((UDPContext *)m_client)->idx, (struct sockaddr *)&addr) < 0)
				err("cannot join multicast group %s !\n", mcg[i]);
		}
	}
	for (int cam=0; cam<NumCam; cam++) m_cam[cam] = new Cam(cam);
	for (int tuner=0; tuner<NumTuner; tuner++) m_tuner[tuner] = new Tuner(tuner);
}
Netceiver::~Netceiver()
{
	for (int cam=0; cam<NumCam; cam++) delete m_cam[cam];
	for (int tuner=0; tuner<NumTuner; tuner++) delete m_tuner[tuner];
	if (m_client) udp_close(((UDPContext *)m_client));
	delete[] m_buf;
	delete[] m_udp;
}

void Netceiver::run(int loops)
{
	char *line = m_buf;
	for (int loop=0; loops==-1 || loop<loops; )
	{
		int len = udp_read(((UDPContext *)m_client), m_udp, UDP_TX_BUF_SIZE, 50, NULL);
		for (int i=0; i<len; i++)
		{
			if (m_udp[i] != '\n') *line++ = m_udp[i];
			else
			{
				*line++ = 0;
				parseLine(clock(), m_buf);
				line = m_buf;
			}
		}
		loop++;
	}
}

namespace
{
	template<typename T> void changed(clock_t t, Netceiver *nc, const char *name, T value) { strstream n,v; n<<name<<'\0'; v<<value<<'\0'; nc->valueChanged(t, n.str(), v.str()); }
	template<typename T> void changed(clock_t t, Netceiver *nc, const char *name, int id, T value) { strstream n,v; n<<name<<"."<<id<<'\0'; v<<value<<'\0'; nc->valueChanged(t, n.str(), v.str()); }
	template<typename T> void changed(clock_t t, Netceiver *nc, const char *name, int id, const char *type, T value) { strstream n,v; n<<name<<"."<<id<<"."<<type<<'\0'; v<<value<<'\0'; nc->valueChanged(t, n.str(), v.str()); }
}

void Netceiver::parseLine(clock_t t, char *buf)
{
	if (!buf || !*buf) return;
	while (isspace(*buf)) buf++;
	if (!strncmp(buf, "slot ", 5)) setSlot(t, &buf[5]);
	else if (!strncmp(buf, "MLD MCGs", 8)) setMCGs(t, &buf[8]);
	else if (!strncmp(buf, "00.CAM status", 13)) setCam(t, 0, &buf[13]);
	else if (!strncmp(buf, "01.CAM status", 13)) setCam(t, 1, &buf[13]);
	else if (!strncmp(buf, "+++", 3)) setXYZ(t, '+', &buf[3]);
	else if (!strncmp(buf, "---", 3)) setXYZ(t, '-', &buf[3]);
	else if (!strncmp(buf, "###", 3)) setXYZ(t, '#', &buf[3]);
	else unexpected(t, buf);
}

char *Netceiver::prep(char *buf)
{
	if (*buf==':') buf++;
	while (isspace(*buf)) buf++;
	while (isspace(buf[strlen(buf)-1])) buf[strlen(buf)-1] = 0;
	return buf;
}

void Netceiver::setSlot(clock_t t, char *buf)
{
	buf = prep(buf);
	int num, type;
	if (buf && sscanf(buf, "%d/%d", &num, &type) == 2)
	{
		if (num<0 || num>NumTuner-1) return;
		Tuner *tuner = m_tuner[num];
		beforeUpdate(t, tuner);
		bool first = (tuner->type != type ? true : false);
		if (tuner->type != type) changed(t, this, "tuner", num, "type", type);
		tuner->type = type;
		buf = strstr(buf, "|");
		if (buf) buf = prep(++buf);
		int count;
		if (buf && sscanf(buf, "count %d", &count) == 1)
		{
			if (first || tuner->count != count) changed(t, this, "tuner", num, "count", count);
			tuner->count = count;
			buf = strstr(buf, "|");
			if (buf) buf = prep(++buf);
			int freq;
			if (buf && sscanf(buf, "freq %d", &freq) == 1)
			{
				if (first || tuner->freq != freq) changed(t, this, "tuner", num, "freq", freq);
				tuner->freq = freq;
				buf = strstr(buf, "|");
				if (buf) buf = prep(++buf);
				int status;
				if (buf && sscanf(buf, "status %02x", &status) == 1)
				{
					if (first || tuner->status != status) changed(t, this, "tuner", num, "status", status);
					tuner->status = status;
					buf = strstr(buf, "|");
					if (buf) buf = prep(++buf);
					char *buf2 = strstr(buf, "|");
					if (buf2)
					{
						buf[buf2-buf-1] = 0;
						if (first || tuner->id != buf) { changed(t, this, "tuner", num, "id", buf); tuner->id = buf; }
						buf = prep(++buf2);
						unsigned int signal;
						if (buf && sscanf(buf, "signal %04x", &signal) == 1)
						{
							if (first || tuner->signal != signal) changed(t, this, "tuner", num, "signal", signal);
							tuner->signal = signal;
							buf = strstr(buf, "|");
							if (buf) buf = prep(++buf);
							unsigned int snr;
							if (buf && sscanf(buf, "snr %04x", &snr) == 1)
							{
								if (first || tuner->snr != snr) changed(t, this, "tuner", num, "snr", snr);
								tuner->snr = snr;
								buf = strstr(buf, "|");
								if (buf) buf = prep(++buf);
								unsigned int ber;
								if (buf && sscanf(buf, "ber %08x", &ber) == 1)
								{
									if (first || tuner->ber != ber) changed(t, this, "tuner", num, "ber", ber);
									tuner->ber = ber;
									buf = strstr(buf, "|");
									if (buf) buf = prep(++buf);
									unsigned int unc;
									if (buf && sscanf(buf, "unc %08x", &unc) == 1)
									{
										if (first || tuner->unc != unc) changed(t, this, "tuner", num, "unc", unc);
										tuner->unc = unc;
										buf = strstr(buf, "|");
										if (buf)
										{
											buf = prep(++buf);
											if (!strncmp(buf, "LOCK", 4))
											{
												if (first || !tuner->lock) changed(t, this, "tuner", num, "lock", true);
												tuner->lock = true;
											}
											else
											{
												if (first || tuner->lock) changed(t, this, "tuner", num, "lock", false);
												tuner->lock = false;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		afterUpdate(t, tuner);
	}
}

void Netceiver::setMCGs(clock_t t, char *buf)
{
	beforeUpdate(t);
	buf = prep(buf);
	int mldmcgs=m_mldmcgs, subscribers=m_subscribers;
	if (sscanf(buf, "%d Subscribers: %d", &m_mldmcgs, &m_subscribers) == 2)
	{
		if (mldmcgs != m_mldmcgs) changed(t, this, "MLD MCGs", m_mldmcgs);
		if (subscribers != m_subscribers) changed(t, this, "Subscribers", m_subscribers);
	}
	afterUpdate(t);
}

void Netceiver::setCam(clock_t t, int cam, char *buf)
{
	beforeUpdate(t, m_cam[cam]);
	buf = prep(buf);
	//(MMI MSG NO): READY (STACK : OK BUSY : 0) (MMI HDL/STA: 0/0/0) (TL: 0x204056d0/1) (D/T: 1/1337633873) TC_POLL failed: 0/0 Flags: 2/0x100/4
//MMI MSG NO
//READY
//STACK : OK BUSY : 0
//MMI HDL/STA: 0/0/0
//TL: 0x204056d0/1
//D/T: 1/1337633873
//TC_POLL failed: 0/0
//Flags: 2/0x100/4
	if (cam==0 && m_cam[cam]->setStatus(buf)) changed(t, this, "cam", cam, m_cam[cam]->getStatus());
	else if (cam==1 && m_cam[cam]->setStatus(buf)) changed(t, this, "cam", cam, m_cam[cam]->getStatus());
	afterUpdate(t, m_cam[cam]);
}

void Netceiver::setXYZ(clock_t t, char type, char *buf) { unexpected(t, buf-3); }
