#ifndef _MCLI_H_
#define _MCLI_H_

#ifndef DllExport
#define DllExport __declspec(dllimport)
#endif

#include <time.h>
#include <string>
using namespace std;

// -----------------------------------------------------------------------------

#define NumTuner 6
#define NumCam 2

DllExport string format(const char *fmt, ...);

// -----------------------------------------------------------------------------

class Netceiver
{
public:
	DllExport Netceiver(const string &ifname="*", int port=23000, char mcg[][1024]=NULL, int num_mcg=1);
	DllExport virtual ~Netceiver();

	DllExport void run(int loops=-1);

	class Cam
	{
	public:
		Cam(int id) : m_id(id) {}
		bool isInstalled() const { return getStatus()=="MISSING"?false:true; }
		int getNum() const { return m_id; }
		string getName() const { return format("Cam %d", 1+getNum()); }
		const string &getStatus() const { return m_status; }
		bool setStatus(const string &buf) { if (m_status == buf) return false; m_status = buf; return true; }
	private:
		int m_id;
		string m_status;
	};
	Cam *getCam(int i) { return m_cam[i]; }

	class Tuner
	{
	public:
		Tuner(int id) : m_id(id), type(-1), count(-1), freq(-1), status(-1), signal(0xffffffff), snr(0xffffffff), ber(0xffffffff), unc(0xffffffff), lock(false) {}
		bool isInstalled() const { return getType()=="MISSING" ? false:true; }
		bool isOff() const { return count==0 && ber==0xdeaddead && unc==0xdeaddead ? true:false; }
		int getNum() const { return m_id; }
		string getName() { return format("Tuner %d (%s)", 1+getNum(), getType().c_str()); }
		string getType() const
		{
			switch (type)
			{
				case 0: return "DVB-S";
				case 1: return "DVB-C";
				case 2: return "DVB-T";
				case 4: return "DVB-S2";
			}
			return "MISSING";
		}
	//private:
		int m_id;
		friend class Netceiver;
		int type, count, freq, status;
		string id;
		unsigned int signal, snr, ber, unc;
		bool lock;
	};
	Tuner *getTuner(int i) { return m_tuner[i]; }

	virtual void valueChanged(clock_t t, const string &name, const string &value) { }
	virtual void beforeUpdate(clock_t t) {}
	virtual void afterUpdate(clock_t t) {}
	virtual void beforeUpdate(clock_t t, Tuner *tuner) {}
	virtual void afterUpdate(clock_t t, Tuner *tuner) {}
	virtual void beforeUpdate(clock_t t, Cam *cam) {}
	virtual void afterUpdate(clock_t t, Cam *cam) {}

private:
	string m_ifname;
	char m_mcg[10][1024];
	int m_num_mcg, m_port, m_mldmcgs, m_subscribers;
	void *m_client;
	Cam *m_cam[NumCam];
	Tuner *m_tuner[NumTuner];
	unsigned char *m_udp;
	char *m_buf;
	void parseLine(clock_t t, char *buf);
	char *prep(char *buf);
	void setSlot(clock_t t, char *buf);
	void setMCGs(clock_t t, char *buf);
	void setCam(clock_t t, int cam, char *buf);
	void setXYZ(clock_t t, char type, char *buf);
	virtual void unexpected(clock_t t, const char *line) {}
};

// -----------------------------------------------------------------------------

extern "C" DllExport bool simRepeat;
extern "C" DllExport bool simSlow;
extern "C" DllExport void setSimFile(const char *f);

// -----------------------------------------------------------------------------

#endif
