#pragma once

#include "resource.h"

#include "c:\temp\netceiver\mcli\mcli.h"
#pragma comment(lib, "c:\\temp\\netceiver\\Debug\\mcli.lib")
#pragma comment(lib, "ws2_32")
#pragma comment(lib, "iphlpapi")

#include "window.h"

extern const char *simFile;

void unexpected(clock_t t, const char *line);

void showTuner(int t=-1);



class Page
{
public:
	Page(GlutWindow& window) : m_window(window), m_mode(SIGNAL)
	{
		assert(_this() == NULL);
		_this() = this;
		m_menu = glutCreateMenu(_menu);
		glutAttachMenu(GLUT_RIGHT_BUTTON);
		glutAddMenuEntry("> quit", 999);
	}
	virtual ~Page() = 0
	{
		glutDetachMenu(GLUT_RIGHT_BUTTON);
		glutDestroyMenu(m_menu);
		_this() = NULL;
	}
	virtual void display(void) {}
	virtual void update(clock_t t, Netceiver::Tuner *tuner) {}
	virtual void update(clock_t t, Netceiver::Cam *cam) {}
	virtual void key(GLubyte key, GLint x, GLint y) {}
protected:
	GlutWindow& m_window;
	enum Mode { SIGNAL, SNR, BER, UNC, COUNT, FREQ, STATUS, LOCK } m_mode;
	double quality(Netceiver::Tuner *tuner, Mode mode, bool raw=false)
	{
		switch (mode)
		{
			case SIGNAL: return raw ? tuner->signal : 100.0*(tuner->signal/65535.0);
			case SNR: return raw ? tuner->snr : 100.0*(tuner->snr/65535.0);
			case BER: return raw ? tuner->ber : 100.0-100.0*(tuner->ber/65535.0);
			case UNC: return raw ? tuner->unc : 100.0-100.0*(tuner->unc/65535.0);
			case COUNT: return raw ? tuner->count : tuner->count*2;//count-max=50?
			case FREQ: return raw ? tuner->freq : (tuner->freq)/20000;//0..2000000 -> 0..100
			case STATUS: return raw ? tuner->status : tuner->status*3;//0..0x1f -> 0..<100
			case LOCK: return raw ? tuner->lock : tuner->lock?100:0;
		}
		return 0;
	}
private:
	int m_menu;
	static Page *&_this() { static Page *rc = NULL; return rc; }
	static void _menu(int item) { if (item == 999) exit(0); _this()->menu(item); }
	virtual void menu(int item) {}
};



class AllTunerPage : public Page
{
public:
	AllTunerPage(GlutWindow& window) : Page(window)
	{
//		clock_t t = clock();
		for (int a=0; a<NumCam; a++)
		{
			m_textCam[a] = new GlutTextControl(m_window, a==0?-m_window.m_w/2:0, m_window.m_h/2-(100*(5+1))-10, "Cam %d", a+1);
//			afterUpdate(t, getCam(a));
		}
		for (int a=0; a<NumTuner; a++)
		{
			m_graph[a] = new GlutGraphControl(m_window, -m_window.m_w/2, m_window.m_h/2-(100*(a+1)), m_window.m_w);
			m_text[a] = new GlutTextControl(m_window, -m_window.m_w/2, m_window.m_h/2-(10+100*a), "%d", a+1);
			m_textTunerType[a] = new GlutTextControl(m_window, m_window.m_w/2-270, m_window.m_h/2-(10+100*a), "<not installed>");
//			afterUpdate(t, getTuner(a));
		}
		#define Q(x,y) glutAddMenuEntry("show "#y" quality", Page::x);
		Q(SIGNAL, signal) Q(SNR, snr) Q(BER, ber) Q(UNC, unc) Q(COUNT, count) Q(FREQ, freq) Q(STATUS, status) Q(LOCK, lock)
		#undef Q
		glutAddMenuEntry("> tuner 1", 100);
		glutAddMenuEntry("> tuner 2", 101);
		glutAddMenuEntry("> tuner 3", 102);
		glutAddMenuEntry("> tuner 4", 103);
		glutAddMenuEntry("> tuner 5", 104);
		glutAddMenuEntry("> tuner 6", 105);
	}
	~AllTunerPage()
	{
		for (int a=0; a<NumCam; a++) delete[] m_textCam[a];
		for (int a=0; a<NumTuner; a++)
		{
			delete[] m_textTunerType[a];
			delete[] m_text[a];
			delete[] m_graph[a];
		}
	}
	void display(void)
	{
		string m = "???";
		switch (m_mode)
		{
			#define Q(x,y) case Page::x: m=#y; break;
			Q(SIGNAL, signal) Q(SNR, snr) Q(BER, ber) Q(UNC, unc) Q(COUNT, count) Q(FREQ, freq) Q(STATUS, status) Q(LOCK, lock)
			#undef Q
		}
		glColor3f(1, 1, 0);
		m_window.print(m_window.m_w/2-70, -m_window.m_h/2+10, "mode: %s", m.c_str());
	}
	void update(clock_t t, Netceiver::Tuner *tuner)
	{
		GlutTextControl *tunerTypeControl = m_textTunerType[tuner->getNum()];
if (!tuner->isInstalled()) tunerTypeControl->text(1,0,0, "");
else tunerTypeControl->text(0,1,0, "%s", tuner->isInstalled() ? (tuner->getType()+"        "+tuner->id).c_str() : "---");
		GlutTextControl *tt = m_text[tuner->getNum()];
		GlutGraphControl *tg = m_graph[tuner->getNum()];
		if (!tuner->isInstalled()) { tt->text(1, 0, 0, "tuner %d is not installed", tuner->getNum()); tg->push(-1); return; }
		if (tuner->isOff()) { tt->text(0.5, 0.5, 1, "OFF"); tg->push(-10);return; }
		double q = quality(tuner, m_mode);
		tg->push(tuner->isOff() ? -10 : q*.9);
#define FMT "sgl:%.1lf   snr:%.1lf   ber:%.1lf   unc:%.1lf   cnt:%d   frq:%.3lfMhz   sts:0x%02x   lck:%s   (%s)"
#define TQ(x) , quality(tuner, x)
#define TC , tuner->count
#define TF , tuner->freq/1000.0
#define TS , tuner->status
#define TL , tuner->lock?"true":"false"
		if (!tuner->count) tt->text(1, 1, 0, FMT TQ(SIGNAL) TQ(SNR) TQ(BER) TQ(UNC) TC TF TS TL, tuner->lock?"no count":"no count / NO LOCK");
		else if (!tuner->lock) tt->text(1, 0.5, 0.5, FMT TQ(SIGNAL) TQ(SNR) TQ(BER) TQ(UNC) TC TF TS TL, "NO LOCK");
		else if (q>=80) tt->text(0, 1, 0, FMT TQ(SIGNAL) TQ(SNR) TQ(BER) TQ(UNC) TC TF TS TL, "QUALITY OK");
		else if (q>=60) tt->text(1, 1, 0, FMT TQ(SIGNAL) TQ(SNR) TQ(BER) TQ(UNC) TC TF TS TL, "QUALITY < 80");
		else tt->text(1, 0, 0, FMT TQ(SIGNAL) TQ(SNR) TQ(BER) TQ(UNC) TC TF TS TL, "QUALITY < 60");
#undef FMT
#undef TQ
#undef TC
#undef TF
#undef TS
#undef TL
	}
	void update(clock_t t, Netceiver::Cam *cam)
	{
		GlutTextControl *ct = m_textCam[cam->getNum()];
		if (!cam->isInstalled()) { ct->text(1, 0, 0, "cam %d is not installed", cam->getNum()); return; }
		string s = cam->getStatus();
		for (int p=s.find(" ("); p!=string::npos; p=s.find(" (", p+1)) s = s.replace(p, 2, "\n(");
		for (int p=s.find(") "); p!=string::npos; p=s.find(") ", p+1)) s = s.replace(p, 2, ")\n");
		int r = s.find("): READY");
		int o = s.find("(STACK : OK");
		int f = s.find("TC_POLL failed: 0/0");
		int b = s.find("BUSY : 1)");
		if (r < 0) ct->text(1,0,0, s.c_str());
		else if (o < 0) ct->text(1,0,0, s.c_str());
		else if (f < 0) ct->text(1,0,0, s.c_str());
		else if (b > 0) ct->text(1,1,0, s.c_str());
		else ct->text(0,0.5,1, s.c_str());
		if (r<0 || o<0 || f<0) unexpected(t, s.c_str());
	}
	void key(GLubyte key, GLint x, GLint y) { if (key>='1' && key<='9') m_mode = (Mode)(key-'1'+SIGNAL); }
private:
	GlutTextControl *m_text[NumTuner], *m_textTunerType[NumTuner];
	GlutGraphControl *m_graph[NumTuner];
	GlutTextControl *m_textCam[NumCam];
	void menu(int item) { if (item < 100) m_mode = (Mode)item; else showTuner(item-100); }
};



class SingleTunerPage : public Page
{
public:
	SingleTunerPage(GlutWindow& window, Netceiver::Tuner *tuner) : Page(window), m_tuner(tuner), m_cam(NULL) { init(); }
	SingleTunerPage(GlutWindow& window, Netceiver::Cam *cam) : Page(window), m_tuner(NULL), m_cam(cam){ init(); }
	~SingleTunerPage()
	{
		#define Q(x,y) delete y; delete text[x];
		Q(SIGNAL, signal) Q(SNR, snr) Q(BER, ber) Q(UNC, unc) Q(COUNT, count) Q(FREQ, freq) Q(STATUS, status) Q(LOCK, lock)
		#undef Q
	}
	void display(void)
	{
		glColor3f(1, 1, 0);
		if (m_tuner) m_window.print(m_window.m_w/2-70, -m_window.m_h/2+10, "tuner: %d", m_tuner->getNum()+1);
		if (m_cam) m_window.print(m_window.m_w/2-70, -m_window.m_h/2+10, "cam: %d", m_cam->getNum()+1);
	}
	void update(clock_t t, Netceiver::Tuner *tuner)
	{
		if (tuner != m_tuner) return;
		#define Q1(x,y) y->push(quality(tuner, x)); text[x]->text(#y": %.2lf (%.0lf)", quality(tuner, x), quality(tuner, x, true));
		#define Q2(x,y) y->push(quality(tuner, x)); text[x]->text(#y": %.0lf", quality(tuner, x, true));
		Q1(SIGNAL, signal) Q1(SNR, snr) Q1(BER, ber) Q1(UNC, unc) Q2(COUNT, count) Q2(FREQ, freq) Q2(STATUS, status) Q2(LOCK, lock)
		#undef Q
	}
	void key(GLubyte key, GLint x, GLint y) { if (key>='1' && key<='9') showTuner(); }
Netceiver::Tuner *m_tuner;
Netceiver::Cam *m_cam;
private:
	GlutGraphControl *signal, *snr, *ber, *unc, *count, *freq, *status, *lock;
	GlutTextControl *text[8];
	void init()
	{
		#define Q(x,y) y = new GlutGraphControl(m_window, -m_window.m_w/2, m_window.m_h/2-(x+1)*100, m_window.m_w);
		Q(SIGNAL, signal) Q(SNR, snr) Q(BER, ber) Q(UNC, unc) Q(COUNT, count) Q(FREQ, freq) Q(STATUS, status) Q(LOCK, lock)
		#undef Q
for (int a=0; a<8; a++) text[a] = new GlutTextControl(m_window, -m_window.m_w/2, m_window.m_h/2-(100*(a+1)-90), "");
glutAddMenuEntry("show all tuner", 1);
glutAddMenuEntry("> tuner 1", 100);
glutAddMenuEntry("> tuner 2", 101);
glutAddMenuEntry("> tuner 3", 102);
glutAddMenuEntry("> tuner 4", 103);
glutAddMenuEntry("> tuner 5", 104);
glutAddMenuEntry("> tuner 6", 105);
	}
	void menu(int item) { if (item >=100) showTuner(item-100); else showTuner(); }
};



class Lock
{
public:
	Lock() { InitializeCriticalSection(&m_cs);(&m_cs); }
	~Lock() { DeleteCriticalSection(&m_cs); }
	void enter() { EnterCriticalSection(&m_cs); }
	void leave() { LeaveCriticalSection(&m_cs); }
	class L
	{
	public:
		L(Lock& lock) : m_lock(lock) { m_lock.enter(); }
		~L() { m_lock.leave(); }
	private:
		Lock& m_lock;
	};
private:
	CRITICAL_SECTION m_cs;
};



class NetceiverWindow : public Netceiver, public GlutWindow
{
public:
	#pragma warning(push)
	#pragma warning(disable: 4355)
	NetceiverWindow(const string &ifname="*", int port=23000, char mcg[][1024]=NULL, int num_mcg=1, const char *title="NetCeiver", int w=800, int h=800)
		: Netceiver(ifname, port, mcg, num_mcg), GlutWindow(title, w, h), m_allTunerPage(NULL), m_singleTunerPage(NULL)
, m_cube(*this, 0, 0, 250)
	{
m_cube.enabled=false;
		clock_t t = clock();
		m_allTunerPage = new AllTunerPage(*this);
	}
	#pragma warning(pop)
	~NetceiverWindow()
	{
		delete m_allTunerPage;
		delete m_singleTunerPage;
	}

	void showAll()
	{
		Lock::L lock(m_lock);
		delete m_singleTunerPage;
		m_singleTunerPage = NULL;
		if (!m_allTunerPage) m_allTunerPage = new AllTunerPage(*this);
	}

	void showTuner(int n)
	{
		Lock::L lock(m_lock);
		delete m_allTunerPage;
		m_allTunerPage = NULL;
		Tuner *tuner = getTuner(n);
		if (m_singleTunerPage && tuner==m_singleTunerPage->m_tuner) return;
		delete m_singleTunerPage;
		m_singleTunerPage = new SingleTunerPage(*this, tuner);
	}

	void showCam(int n)
	{
		Lock::L lock(m_lock);
		delete m_allTunerPage;
		m_allTunerPage = NULL;
		Cam *cam = getCam(n);
		if (m_singleTunerPage && cam==m_singleTunerPage->m_cam) return;
		delete m_singleTunerPage;
		m_singleTunerPage = new SingleTunerPage(*this, cam);
	}

	void run()
	{
		HANDLE t = CreateThread(NULL, 0, receiver, this, 0, NULL);
		glutMainLoop();
		TerminateThread(t, 0);
	}

private:
GlutCubeControl m_cube;
	Lock m_lock;
	AllTunerPage *m_allTunerPage;
	SingleTunerPage *m_singleTunerPage;
	Page *currentPage()
	{
		if (m_allTunerPage) return m_allTunerPage;
		if (m_singleTunerPage) return m_singleTunerPage;
		return NULL;
	}
	static DWORD WINAPI receiver(LPVOID lpThreadParameter) { NetceiverWindow *_this = (NetceiverWindow *)lpThreadParameter; _this->Netceiver::run(); return 0; }
	void idle() { glutPostRedisplay(); }
	void timer(int value) { m_cube.spin(); } // graph[0]->xxx++;
	void display(void) { Lock::L lock(m_lock); currentPage()->display(); }
	void key(GLubyte key, GLint x, GLint y)
	{
		Lock::L lock(m_lock);
		if (key == 'c') m_cube.enabled = 1-m_cube.enabled;
		else if (key == 27) exit(0);
		else currentPage()->key(key, x, y);
		GlutWindow::key(key, x, y);
	}
	void fk(int n)
	{
		Lock::L lock(m_lock);
		if (n>=1 && n<=NumTuner) showTuner(n-1); // f1..f6: show tuner 1..6
		else if (n>=NumTuner+1 && n<=NumTuner+NumCam) showCam(n-NumTuner-1); // f7..f8: show cam 1..2
		else showAll(); // f9...: show all tuner
	}
	void afterUpdate(clock_t t, Tuner *tuner) { Lock::L lock(m_lock); currentPage()->update(t, tuner); }
	void afterUpdate(clock_t t, Cam *cam) { Lock::L lock(m_lock); currentPage()->update(t, cam); }
};
