#ifndef _WINDOW_H_
#define _WINDOW_H_

class GlutControl;

// -----------------------------------------------------------------------------

class GlutWindow
{
public:
	GlutWindow(const string& title, int w=800, int h=600)
	{
		assert(_this() == NULL);
		_this() = this;
		m_x = m_y = 0;
		m_w = w;
		m_h = h;
		char buf1[64*1024] = "exe";
		char buf2[64*1024] = "param";
		size_t l;
		errno_t e = wcstombs_s(&l, buf1, __wargv[0], 64*1024);
		int argc = __argc;
		if (argc>0 && __wargv[1]) e = wcstombs_s(&l, buf2, __wargv[1], 64*1024);
		else argc = 0;
		char *argv[] = { buf1, argc>0?buf2:NULL, NULL };
		glutInit(&argc, argv);
		glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
		glutInitWindowSize(m_w+2, m_h+2);
		glutCreateWindow(title.c_str());
		glColor3f(1, 1, 1);
		glutDisplayFunc(_display);
		glutReshapeFunc(_reshape);
		glutKeyboardFunc(_key);
		glutMouseFunc(_mouseClick);
		glutMotionFunc(_mouseMove);
		glutSpecialFunc(_special);
		glutIdleFunc(_idle);
		glutTimerFunc(500, _timer, 100);
		glutGameModeString("1920x1200:32");
//int menu = glutCreateMenu(_menu);
//glutAddMenuEntry("Quit", 999);
//glutAttachMenu(GLUT_RIGHT_BUTTON);
		//if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) glutEnterGameMode();
		//glEnable(GL_DEPTH_TEST);
	}
	virtual ~GlutWindow() = 0 {}

	void print(int x, int y, const char *fmt, ...)
	{
		va_list args;
		char buf[64*1024], *s=buf;
		va_start(args, fmt);
		vsprintf_s(buf, fmt, args); // vsprintf(buf, fmt, args);
		glRasterPos2d(x, y);
		while (*s)
		{
			if (*s == '\n') { y -= 10; glRasterPos2d(x, y); s++; }
			else glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *s++);//GLUT_BITMAP_TIMES_ROMAN_10
		}
	}
	void line(int x1, int y1, int x2, int y2)
	{
		glBegin(GL_LINES);
		glVertex2d(x1, y1);
		glVertex2d(x2, y2);
		glEnd();
	}
	void rect(int x1, int y1, int x2, int y2)
	{
		glBegin(GL_QUADS);
		glVertex2d(x1, y1);
		glVertex2d(x2, y1);
		glVertex2d(x2, y2);
		glVertex2d(x1, y2);
		glEnd();
	}
	void graph(int num, int x[], int y[])
	{
		glBegin(GL_QUAD_STRIP);
		for (int a=0; a<num; a++)
		{
			if (y[a] < 0) glColor3f(0, 0, 1);
			else glColor3f(1.0-y[a]/100.0, y[a]/100.0, 0);
			glVertex2d(x[a], 0);
			glVertex2d(x[a], abs(y[a]));
		}
		glEnd();
	}

	void run() { glutMainLoop(); }

//protected:
	int m_x, m_y, m_w, m_h;

//private:
	static GlutWindow *&_this() { static GlutWindow *rc = NULL; return rc; }
	static void _display();
	virtual void display(void) {}
	static void _reshape(int w, int h) { _this()->reshape(w, h); }
	virtual void reshape(int w, int h)
	{
		glViewport(0, 0, w, h);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		//glFrustum(-1000.0, 1000.0, -1000.0, 1000.0, -1000.0, 1000.0);
		GLdouble size = (GLdouble)((w >= h) ? w : h) / 2.0, aspect;
		if (w <= h) { aspect = (GLdouble)h/(GLdouble)w; glOrtho(-size, size, -size*aspect, size*aspect, -100000.0, 100000.0); }
		else { aspect = (GLdouble)w/(GLdouble)h; glOrtho(-size*aspect, size*aspect, -size, size, -100000.0, 100000.0); }
		glScaled(aspect, aspect, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}
	static void _key(GLubyte key, GLint x, GLint y) { _this()->key(key, x, y); }
	virtual void key(GLubyte key, GLint x, GLint y)
	{
		switch (key)
		{
			case 27: exit(0);
			case 'w': m_y++; break;
			case 's': m_y--; break;
			case 'a': m_x--; break;
			case 'd': m_x++; break;
		}
		glutPostRedisplay();
	}
	static void _mouseClick(int button, int state, int x, int y) { _this()->mouseClick(button, state, x, y); }
	virtual void mouseClick(int button, int state, int x, int y) {}
	static void _mouseMove(int x, int y) { _this()->mouseMove(x, y); }
	virtual void mouseMove(int x, int y) { m_x=x; m_y=y; }
	static void _timer(int value) { _this()->timer(value); glutTimerFunc(value, _timer, value); }
	virtual void timer(int value) {}
	//static void _menu(int item) { if (item==999) exit(0); _this()->menu(item); }
	//virtual void menu(int item) {}
	static void _special(int key, GLint x, GLint y) { _this()->special(key, x, y); }
	void special(int key, int x, int y)
	{
		if (key>=GLUT_KEY_F1 && key<=GLUT_KEY_F12) fk(key);
		switch (key)
		{
			case GLUT_KEY_UP: m_y++; break;
			case GLUT_KEY_DOWN: m_y--; break;
			case GLUT_KEY_LEFT: m_x--; break;
			case GLUT_KEY_RIGHT: m_x++; break;
		}
		glutPostRedisplay();
	}
	virtual void fk(int n) {}
	static void _idle() { _this()->idle(); }
	virtual void idle() {}
	friend class GlutControl;
	list<GlutControl *> m_controls;
	void add(GlutControl& control) { m_controls.push_back(&control); }
	void remove(GlutControl& control) { m_controls.remove(&control); }
};

// -----------------------------------------------------------------------------

class GlutControl
{
public:
	GlutControl(GlutWindow &window, int x, int y) : m_window(window), m_x(x), m_y(y), enabled(true) { m_window.add(*this); }
	virtual ~GlutControl() = 0 { m_window.remove(*this); }
	bool enabled;
protected:
	GlutWindow &m_window;
	int m_x, m_y;
	friend class GlutWindow;
	virtual void draw() = 0 {}
};

// -----------------------------------------------------------------------------

void GlutWindow::_display()
{
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	for(list<GlutControl *>::iterator i=_this()->m_controls.begin(); i!=_this()->m_controls.end(); i++) (*i)->draw();
	_this()->display();
	glFlush();
	glutSwapBuffers();
}

// -----------------------------------------------------------------------------

class GlutTextControl : public GlutControl
{
public:
	GlutTextControl(GlutWindow &window, int x, int y, const char *fmt, ...) : GlutControl(window, x, y), m_r(1), m_g(1), m_b(1)
	{ va_list args; char buf[64*1024]; va_start(args, fmt); vsprintf_s(buf, fmt, args); m_text = buf; }
	~GlutTextControl() {}
	void text(const char *fmt, ...) { m_r=m_g=m_b=1; va_list args; char buf[64*1024]; va_start(args, fmt); vsprintf_s(buf, fmt, args); m_text = buf; }
	void text(float r, float g, float b, const char *fmt, ...) { m_r=r; m_g=g; m_b=b; va_list args; char buf[64*1024]; va_start(args, fmt); vsprintf_s(buf, fmt, args); m_text = buf; }
private:
	float m_r, m_g, m_b;
	string m_text;
	void draw() { if (!enabled) return; glColor3f(m_r, m_g, m_b); m_window.print(m_x, m_y, m_text.c_str()); }
};

// -----------------------------------------------------------------------------

class GlutGraphControl : public GlutControl
{
public:
	GlutGraphControl(GlutWindow &window, int x, int y, int num) : GlutControl(window, x, y), m_num(num), m_pos(0), m_px(new int[m_num]), m_py(new int[m_num])
,xxx(0)
	{ for (int a=0; a<m_num; a++) { m_px[a] = a; m_py[a] = 1; } }
	~GlutGraphControl() { delete[] m_px; delete[] m_py; }
	void push(int y) { m_pos=m_pos%m_num; m_py[m_pos++]=y; }
int xxx;
private:
	int m_num, m_pos;
	int *m_px, *m_py;
	void draw()
	{
		if (!enabled) return;
		glPushMatrix();
glRotated(xxx, 1, 1, 1);
		glTranslated(m_x, m_y, 0);
		m_window.graph(m_num, m_px, m_py);
		glColor3f(1, 1, 0);
		glBegin(GL_QUADS);
		glVertex2d(m_px[m_pos-1]-1, 0);
		glVertex2d(m_px[m_pos-1]-1, 100);
		glVertex2d(m_px[m_pos-1], 100);
		glVertex2d(m_px[m_pos-1], 0);
		glEnd();
		glPopMatrix();
	}
};

// -----------------------------------------------------------------------------

class GlutCubeControl : public GlutControl
{
public:
	GlutCubeControl(GlutWindow &window, int x, int y, int s) : GlutControl(window, x, y), m_s(s), m_a(0) {}
	~GlutCubeControl() {}
	void spin() { m_a++; }
private:
	int m_s, m_a;
	void draw() { if (!enabled) return; glColor3f(0.2, 0.1, 0); glPushMatrix(); glRotated(m_a, 1, 2, 3); glTranslated(m_x, m_y, 0); glutWireCube(m_s); glPopMatrix(); }
};

// -----------------------------------------------------------------------------

#endif
